package RPM::Source::Transformation::Factory::Batch;

use strict;
use warnings;
use Carp;

use Text::ParseWords;
use RPM::Source::Transformation::Factory;
use RPM::Source::Transformation::Factory::Buildreq 0.002;
use RPM::Source::Transformation::Factory::Hook 0.004;
$RPM::Source::Transformation::Factory::Hook::opt_hook_files_for_package=1;
$RPM::Source::Transformation::Factory::Buildreq::opt_buildreq_merge_sourcedep_entry=1;

our @ISA=qw/RPM::Source::Transformation::Factory/;
push @RPM::Source::TransformContainer::FACTORY_CLASSES, __PACKAGE__;

our $VERSION=0.011;
our $verbose=0;

our @opt_batchfiles;
my @opt_oldbatchfiles;
my $opt_rejected_batches_file;
my $opt_ignore_bad_plugin_options;

our @LONGOPT=(
    "batch=s"   => \@opt_batchfiles,
    "oldbatch|old-batch=s"   => \@opt_oldbatchfiles,
    "rejected-batch|already-applied-batch=s"   => \$opt_rejected_batches_file,
    "ignore-bad-plugin-options!"   => \$opt_ignore_bad_plugin_options,
    );

sub print_longopt {
    print "  Logoved batch options:
    --batch <logoved batch file>
    --old-batch <logoved batch file> already applied batch commands
    --rejected-batch <filename> file to write rejected (already appled, 
      but proposed again) batch commands.
";
}

our $PLUGIN={
    'buildreq' => \&plugin_buildreq,
    'buildrequires' => \&plugin_buildreq,
    'perlfixer' => \&plugin_perlfixer,
    'hook' => \&plugin_hook,
};

my $hook_opts={
    '--env' => \&plugin_hook_opt_env,
    '--environment' => \&plugin_hook_opt_env,
};

my $buildreq_opts={
    '--add' => \&plugin_buildreq_opt_add,
    '--add-pre' => \&plugin_buildreq_opt_add_pre,
#    '--add-buildreq' => \&plugin_buildreq_opt_add,
#    '--add-buildreq-pre' => \&plugin_buildreq_opt_add_pre,
};

our $name2batch={};
our $name2nvr={};

my $name2oldbatch={};
my $in_name2batch={};

sub register_plugin {
    my ($name,$coderef)=@_;
    $PLUGIN->{$name}=$coderef;
}


sub process_options {
    if (not @opt_batchfiles and -e 'LOGS/00FIXSCRIPT') {
	warn "no --batch <file> given, but LOGS/00FIXSCRIPT found. loading...\n";
	push @opt_batchfiles, 'LOGS/00FIXSCRIPT';
    }
    map {&load_batchfile_with(\&old_batchfile_handler, $_)} @opt_oldbatchfiles;
    map {&load_batchfile_with(\&new_batchfile_handler, $_)} @opt_batchfiles;
    my @already_applied;
    foreach my $name (keys(%$in_name2batch)) {
	my $new_commands_found=0;
	my $old_commands=$name2oldbatch->{$name};
	my $in_commands=$in_name2batch->{$name};
	foreach my $command (@$in_commands) {
	    if (not $old_commands->{$command}) {
		$new_commands_found=1;
		last;
	    }
	}
	if ($new_commands_found) {
	    push @{$name2batch->{$name}}, @$in_commands;
	    warn 'DEBUG: Batch: ', $name, ': added commands: ',join('<@@>',@$in_commands),"\n" if $verbose>2;
	} else {
	    warn 'WARNING: commands already_applied for ',$name,"\n";
	    push @already_applied, map { $name."\t\t".$_."\n"} @$in_commands;
	}
    }
    # TODO: write if option and @already_applied
    if ($opt_rejected_batches_file and @already_applied) {
	open (my $fn, '>', $opt_rejected_batches_file) || die "failed to write to $opt_rejected_batches_file: $!";
	print $fn @already_applied;
	close($fn);
    }
}

sub register_messages {
    my ($self,$extopt)=@_;
    ### TODO: use some fixed keys hash here
    RPM::Source::Transformation::Component::add_opt($extopt, {
	-batch => undef,
    });
}

sub transformations {
    return RPM::Source::Transformation::OptHandlerWrapper->new(
	-priority => PRIORITY_AT_BEGIN,
	-name => 'batch jobs',
	-opt_handler_sub=>\&opt_handler);
}

sub opt_handler {
    my (undef,$spec,undef,$opt,$resourceobj)=@_;
    my @batch;

    # TODO: make a method in a resource responsible for nvr
    my $rpmname=$opt->{-input_handler_arg_input_name};
    $rpmname//=$spec->macros->macro_subst('%name');

    foreach my $batchref ($name2batch->{$rpmname}, $opt->{-batch}) {
	push @batch, @$batchref if $batchref;
    }
    warn 'DEBUG: Batch: opt_handler: ', $rpmname, ': got batch: ',join('<@@>',@batch),"\n" if $verbose>2;
    foreach my $cmdline (@batch) {
	&apply_batch_job($spec,$opt,$resourceobj,$cmdline)
	    || return return RPM::Source::TransformContainer::TRANSFORMATION_FAILURE;
    }
    return RPM::Source::TransformContainer::TRANSFORMATION_OK;
}

sub load_batchfile_with {
    my ($handler,$batchfile)=@_;
    die 'Oops' unless ref $handler;
    open (my $fh, '<', $batchfile) || die "Can't open ",$batchfile,': ', $!;
    while (my $batchline=<$fh>) {
	chomp $batchline;
	next if $batchline=~/^\s*(?:#|$)/;
	my ($name,$nvr,$cmd) = split("\t", $batchline, 3);
	&$handler( $name,$nvr,$cmd,$batchfile );
    }
    close($fh);
}

sub old_batchfile_handler {
    my ($name,$nvr,$cmd,$batchfile) = @_;
    $name2oldbatch->{$name}->{$cmd}=1;
}

sub new_batchfile_handler {
    my ($name,$nvr,$cmd,$batchfile) = @_;
    my $oldnvr=$name2nvr->{$name};
    $name2nvr->{$name}=$nvr if $nvr;
    die 'batch ',$batchfile,': NVR conflict for ',$name,': ',$oldnvr,' vs ',$nvr,"\n" if $oldnvr and $nvr and $oldnvr ne $nvr;
    push @{$in_name2batch->{$name}}, $cmd;
}

sub apply_batch_job {
    my ($spec,$converter_opt,$resourceobj,$cmdline)=@_;
    $cmdline=~s/\s*$//;
    my $cmd=$cmdline;
    my $argstring='';
    if ($cmdline=~/^(\S+)\s+(\S.*)/) {
	$cmd=$1;
	$argstring=$2;
    }
    my $plugin=$PLUGIN->{$cmd};
    unless ($plugin) {
	die "Batch: unsupported plugin: $cmd";
    }
    return &$plugin($spec,$converter_opt,$resourceobj,$cmd,$argstring);
}

sub plugin_generic {
    my ($spec,$converter_opt,$resourceobj,$cmd,$argstring,$opt_handlers,$arg_handler)=@_;
    unless ($argstring) {
	warn 'Batch: ', $cmd, ": no arguments\n";
	return 0;
    }
    my @argvs=shellwords($argstring);
    if (not $arg_handler and scalar @argvs % 2 !=0) {
	warn 'Batch: ', $cmd, ": odd number of arguments in $argstring\n";
	return 0;
    }
    my @args;
    while (@argvs) {
	my $opt=shift @argvs;
	if ($opt!~/^--/) {
	    if ($arg_handler) {
		push @args, $opt;
	    } elsif ($opt_ignore_bad_plugin_options) {
		warn 'Batch: ', $cmd, ": expected an option, got $opt in $argstring\n";
		return 0;
	    } else {
		die 'Batch: ', $cmd, ": expected an option, got $opt in $argstring
Please, fix the batch file or use --ignore-bad-plugin-options to skip bad lines\n";
	    }
	} else {
	    my $opt_handler=$opt_handlers->{$opt};
	    if (!$opt_handler) {
		if ($opt_ignore_bad_plugin_options) {
		    warn 'Batch: ', $cmd, ": unsupported option $opt in $argstring\n";
		    return 0;
		} else {
		    die 'Batch: ', $cmd, ": unsupported option $opt in $argstring
Please, fix the batch file or use --ignore-bad-plugin-options to skip bad lines\n";
		}
	    }
	    my $optarg=shift @argvs;
	    &$opt_handler($spec,$converter_opt,$resourceobj,$opt,$optarg) || return 0;
	}
    }
    if ($arg_handler) {
	&$arg_handler($spec,$converter_opt,$resourceobj,@args) || return 0;
    }
    return 1;
}

# ---------------------------------------------------

sub plugin_buildreq {
    my ($spec,$converter_opt,$resourceobj,$cmd,$argstring)=@_;
    return &plugin_generic($spec,$converter_opt,$resourceobj,$cmd,$argstring,$buildreq_opts);
}

sub plugin_buildreq_opt_add {
    my ($spec,$converter_opt,$resourceobj,$opt,$arg)=@_;
    $resourceobj->buildreq->add_buildreq_for('batch job',$arg);
    return 1;
}

sub plugin_buildreq_opt_add_pre {
    my ($spec,$converter_opt,$resourceobj,$opt,$arg)=@_;
    $resourceobj->buildreq->add_pre_buildreq_for('batch job',$arg);
    return 1;
}

# ---------------------------------------------------

sub plugin_hook {
    my ($spec,$converter_opt,$resourceobj,$cmd,$argstring)=@_;
    return &plugin_generic($spec,$converter_opt,$resourceobj,$cmd,$argstring,$hook_opts,\&plugin_hook_args);
}

sub plugin_hook_args {
    my ($spec,$converter_opt,$resourceobj,@args)=@_;
    warn 'DEBUG: Batch: plugin_hook_args: set hooks: ',join('<@@>',@args),"\n" if $verbose>2;
    push @{$converter_opt->{-load_hook_files_for_package}}, @args;
}

sub plugin_hook_opt_env {
    my ($spec,$converter_opt,$resourceobj,$opt,$arg)=@_;
    if ($arg=~m!^([\w_]+)=(.*)$!) {
	$converter_opt->{-hooks_environment_for_package}->{$1}=$2;
    } else {
	warn 'hook: bad argument for --$opt: $arg';
    }
    return 1;
}

# ---------------------------------------------------

sub plugin_perlfixer {
    my ($spec,$converter_opt,$resourceobj,$cmd,$argstring)=@_;
    warn "Not implemented: $cmd $argstring\n";
}


1;
